'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

exports.publishable = publishable;
exports.subscribe = subscribe;
exports.unsubscribe = unsubscribe;
exports.publish = publish;
exports.connector = connector;
exports.propConnector = propConnector;
exports.createSubscriber = createSubscriber;

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function specType(spec) {
  return spec === null ? 'null' : typeof spec === 'undefined' ? 'undefined' : _typeof(spec);
}

// Publish/Subscribe implementation

var allSubscriptions = new WeakMap();

var referenceTypes = {
  'object': true,
  'function': true
};

function publishable(subject) {
  return referenceTypes.hasOwnProperty(specType(subject));
};

function subscribe(subject, subscription) {
  var subscriptions = allSubscriptions.get(subject);

  if (subscriptions) {
    subscriptions.push(subscription);
  } else {
    if (publishable(subject)) {
      allSubscriptions.set(subject, [subscription]);
    } else {
      throw new TypeError('subscribe(subject, subscription): subject must be an object or function, got ' + specType(subject));
    }
  }

  return subscription;
};

function unsubscribe(subject, subscriptionToCancel) {
  var subscriptions = allSubscriptions.get(subject);

  if (subscriptions) {
    // Consider computation/memory tradeoff of removing empty subscription lists here
    allSubscriptions.set(subject, subscriptions.filter(function (subscription) {
      return subscription !== subscriptionToCancel;
    }));
  }
};

function publish(subject, value) {
  var subscriptions = allSubscriptions.get(subject);

  if (subscriptions) {
    var numSubscribers = subscriptions.length;
    for (var i = 0; i < numSubscribers; i++) {
      subscriptions[i](value);
    }
  }

  return value;
};

// React-related

var sharedMethods = {
  componentWillMount: function componentWillMount() {
    this.canUpdate = true;
    if (publishable(this.subscription)) {
      subscribe(this.subscription, this.update);
    } // TODO: add a warning when passing unpublishable value to subscription prop?
  },

  componentWillUnmount: function componentWillUnmount() {
    this.canUpdate = false; // prevent updates from triggering re-renders after component decides to unmount
    unsubscribe(this.subscription, this.update);
  },

  update: function update(value) {
    // value is almost always undefined here, but can be because of how zine.publish uses it as an optional second argument
    if (this.canUpdate) {
      this.setState({ props: this._zine.transform(this.subscription, value) });
    }
  },

  render: function render() {
    return _react2.default.createElement(this._zine.component, Object.assign({}, this.props, this.state.props));
  }
};

var staticSubscriberMethods = {
  getInitialState: function getInitialState() {
    this.subscription = this._zine.subscriptionSpec;
    return { props: this._zine.transform(this.subscription) };
  }
};

var dynamicSubscriberMethods = {
  getInitialState: function getInitialState() {
    this.subscription = this.props[this._zine.subscriptionSpec];
    return { props: this._zine.transform(this.subscription) };
  },

  componentWillReceiveProps: function componentWillReceiveProps(nextProps) {
    var oldSubscription = this.props[this._zine.subscriptionSpec];
    var subscription = this.subscription = nextProps[this._zine.subscriptionSpec];
    if (subscription != oldSubscription) {
      unsubscribe(oldSubscription, this.update);
      if (publishable(subscription)) {
        subscribe(subscription, this.update);
      } // TODO: add a warning when passing unpublishable value to subscription prop?
    }
    this.setState({ props: this._zine.transform(subscription) });
  }
};

var typeToMethodMap = {
  'string': dynamicSubscriberMethods,
  'object': staticSubscriberMethods,
  'function': staticSubscriberMethods
};

function pass(sub) {
  // default transform just merges the subscribed object into props
  return sub;
}

function connector(subscriptionSpec) {
  var transform = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : pass;

  var methods = typeToMethodMap[specType(subscriptionSpec)];
  if (!methods) {
    throw new TypeError('connector(subscriptionSpec, transform): subscriptionSpec must be a string, object or function, got ' + specType(subscriptionSpec));
  }
  if (specType(transform) !== 'function') {
    throw new TypeError('connector(subscriptionSpec, transform): transform must be a function, got ' + specType(transform));
  }

  return function (component) {
    return _react2.default.createClass(Object.assign({}, sharedMethods, methods, {
      displayName: 'Subscriber',
      _zine: {
        component: component,
        subscriptionSpec: subscriptionSpec,
        transform: transform
      }
    }));
  };
};

function propConnector(prop) {
  var transform = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : pass;

  if (specType(prop) !== 'string') {
    throw new TypeError('propConnector(prop, transform): prop name must be a string, got ' + specType(prop));
  }
  if (specType(transform) !== 'function') {
    throw new TypeError('propConnector(subscriptionSpec, transform): transform must be a function, got ' + specType(transform));
  }

  return connector(prop, function (sub, val) {
    return _defineProperty({}, prop, transform(sub, val));
  });
};

function createSubscriber(propSpec, component) {
  return (Array.isArray(propSpec) ? propSpec : [propSpec]).reduce(function (comp, prop) {
    return propConnector(prop)(comp);
  }, component);
};

exports.default = { publishable: publishable, subscribe: subscribe, unsubscribe: unsubscribe, publish: publish, connector: connector, propConnector: propConnector, createSubscriber: createSubscriber };
